#!/usr/bin/python
import re
import matplotlib.pyplot as plt

def plotCase(the_file, setTypePattern, label, marker):
    matchPattern = setTypePattern + " (\d+) (.*)"
    with open(the_file, 'r') as f:
        lines = f.readlines()
        x = []
        y = []
        for line in lines:
            theMatch = re.match(matchPattern, line)
            if theMatch != None:
                x.append(float(theMatch.group(1)))
                y.append(float(theMatch.group(2)))
        plt.plot(x, y, label=label,marker=marker)
    

#Ord set scalability

table_types_and_names = [("se.uu.collection.CATreeMapSTDR", "CA-AVL", 'o'),
                         ("se.uu.collection.RangeUpdateLockFreeKSTRQ", "k-ary", 'v'),
                         ("se.uu.collection.FatCATreeMapSTDR", "CA-SL", 'x'),
                         ("se.uu.collection.NonAtomicRangeUpdateConcurrentSkipListMap", "NonAtaomicROSL", '<'),
                         ("se.uu.collection.RangeUpdateSnapTree", "Snap", ">")]

def draw_graph(save_file_name,
               theFile,
               xmax=None,
               ymax=None,
               legendLoc=0):
    plt.figure(figsize=(8,4.75))
    plt.autoscale(enable=True, tight=False)
    if xmax is not None:
    	plt.xlim(0,xmax)
    if ymax is not None:
    	plt.ylim(0,ymax)
    plt.xlabel('Number of Threads')
    plt.ylabel('Operations / Nanosecond') 
    for (name, label, marker) in table_types_and_names:
        plotCase(theFile,
                 name, label, marker)
    plt.legend(loc=legendLoc)
    plt.savefig(save_file_name + '.pdf', bbox_inches='tight')




# draw_graph("java_80_reads",
#            "80 Reads",
#            "\"delete=1048576\"",
#            "", ",1,{0,0,0}\]",
#            2**20,
#            xmax=None,
#            ymax=12,
#            theFile="graph_src/OrdSetScalResults/bench_concops/ets_bench/measurements/R17.PIN_RG.sched.time")

# draw_graph("../../paper_src/ets_scal_insert_bo",
#            "ETS Insert",
#            "\"insert=1048576\"",
#            "", ",1,{0,0,0}\]",
#            2**20,
#            xmax=None,
#            ymax=9,
#            theFile="graph_src/OrdSetScalResults/bench_concops/ets_bench/measurements/R17.PIN_RG.sched.time")


# for (inFile,
#      theType,
#      table_types_and_names) in [
#     ("graph_src/OrdSetScalResults/bench_concops/ets_bench/measurements/R17.PIN_RG.sched.time",
#      "bo", [("\[set,21,20,22,[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?,no", "set", 'o'),
#              ("\[set,21,20,22,[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?,r", "set,r", 'v'),
#              ("\[set,21,20,22,[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?,w", "set,w", '^'),
#              ("\[set,21,20,22,[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?,rw", "set,r,w", '<'),
#              ("\[ordered_set,21,20,22,[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?,no", "ordset", '>'),
#              ("\[ordered_set,21,20,22,[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?,r", "ordset,r", '1')])]:
#     #("plainlib.PIN_RG \[ordered_set,21,20,22,[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?,rw", "ordset,DR-MCS", '>'),
#     #("puredel.PIN_RG \[ordered_set,21,20,22,[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?,rw", "ordset,MR-QD", '1')
#     draw_graph("../../paper_src/ets_scal_lookup_" + theType,
#                "ETS Lookup",
#                "\"lookup=4194304\"",
#                "", ",1,{0,0,0}\]",
#                2**22,
#                xmax=None,
#                ymax=None,
#                legendLoc=0,
#                theFile="graph_src/OrdSetScalResults/bench_concops/ets_bench/measurements/R17.PIN_RG.sched.time")
#     draw_graph("../../paper_src/ets_scal_update_0_" + theType,
#                "ETS Mixed (100% Updates)",
#                "\"mixed=l:0.00%,=u:100.00%=4194304\"",
#                "", ",1,{0,0,0}\]",
#                2**22,
#                xmax=None,
#                ymax=None,
#                legendLoc=0,
#                theFile="graph_src/OrdSetScalResults/bench_concops/ets_bench/measurements/R17.PIN_RG.sched.time")
#     draw_graph("../../paper_src/ets_scal_update_50_" + theType,
#                "ETS Mixed (50% Lookups, 50% Updates)",
#                "\"mixed=l:50.00%,=u:50.00%=4194304\"",
#                "", ",1,{0,0,0}\]",
#                2**22,
#                xmax=None,
#                ymax=None,
#                legendLoc=0,
#                theFile="graph_src/OrdSetScalResults/bench_concops/ets_bench/measurements/R17.PIN_RG.sched.time")
#     draw_graph("../../paper_src/ets_scal_update_80_" + theType,
#                "ETS Mixed (80% Lookups, 20% Updates)",
#                "\"mixed=l:80.00%,=u:20.00%=4194304\"",
#                "", ",1,{0,0,0}\]",
#                2**22,
#                xmax=None,
#                ymax=None,
#                legendLoc=0,
#                theFile="graph_src/OrdSetScalResults/bench_concops/ets_bench/measurements/R17.PIN_RG.sched.time")
#     draw_graph("../../paper_src/ets_scal_update_90_" + theType,
#                "ETS Mixed (90% Lookups, 10% Updates)",
#                "\"mixed=l:90.00%,=u:10.00%=4194304\"",
#                "", ",1,{0,0,0}\]",
#                2**22,
#                xmax=None,
#                ymax=None,
#                legendLoc=0,
#                theFile="graph_src/OrdSetScalResults/bench_concops/ets_bench/measurements/R17.PIN_RG.sched.time")
#     draw_graph("../../paper_src/ets_scal_update_99_" + theType,
#                "ETS Mixed (99% Lookups, 1% Updates)",
#                "\"mixed=l:99.00%,=u:1.00%=4194304\"",
#                "", ",1,{0,0,0}\]",
#                2**22,
#                xmax=None,
#                ymax=None,
#                legendLoc=7,
#                theFile="graph_src/OrdSetScalResults/bench_concops/ets_bench/measurements/R17.PIN_RG.sched.time")



# table_types_and_names = [("set", "set", 'o'),
#                          ("ordered_set", "ordset", 's'),
#                          ("avlatree", "AVL-CA tree", 'D'),
#                          ("treapatree", "Treap-CA tree", '^')]


# #Main Data Structure Benchmark

# draw_graph("../../paper_src/ets_delete",
#            "ETS Delete",
#            "\"delete=1048576\"",
#            "\[", ",21,20,22,1.0,rw,1,{0,0,0}\]",
#            2**20,
#            xmax=None,
#            ymax=12)

# draw_graph("../../paper_src/ets_insert",
#            "ETS Insert",
#            "\"insert=1048576\"",
#            "\[", ",21,20,22,1.0,rw,1,{0,0,0}\]",
#            2**20,
#            xmax=None,
#            ymax=9)

# draw_graph("../../paper_src/ets_lookup",
#            "ETS Lookup",
#            "\"lookup=4194304\"",
#            "\[", ",21,20,22,0,rw,1,{0,0,0}\]",
#            2**22,
#            xmax=None,
#            ymax=None,
#            legendLoc=2)


# draw_graph("../../paper_src/ets_update_0",
#            "ETS Mixed (100% Updates)",
#            "\"mixed=l:0.00%,=u:100.00%=4194304\"",
#            "\[", ",21,20,22,1.0,rw,1,{0,0,0}\]",
#            2**22,
#            xmax=None,
#            ymax=None,
#            legendLoc=1)

# draw_graph("../../paper_src/ets_update_50",
#            "ETS Mixed (50% Lookups, 50% Updates)",
#            "\"mixed=l:50.00%,=u:50.00%=4194304\"",
#            "\[", ",21,20,22,0.5,rw,1,{0,0,0}\]",
#            2**22,
#            xmax=None,
#            ymax=None,
#            legendLoc=4)

# draw_graph("../../paper_src/ets_update_33",
#            "ETS Mixed (33% Lookups, 67% Updates)",
#            "\"mixed=l:33.00%,=u:67.00%=4194304\"",
#            "\[", ",21,20,22,0.67,rw,1,{0,0,0}\]",
#            2**22,
#            xmax=None,
#            ymax=None,
#            legendLoc=4)


# draw_graph("../../paper_src/ets_update_80",
#            "ETS Mixed (80% Lookups, 20% Updates)",
#            "\"mixed=l:80.00%,=u:20.00%=4194304\"",
#            "\[", ",21,20,22,0.2,rw,1,{0,0,0}\]",
#            2**22,
#            xmax=None,
#            ymax=None,
#            legendLoc=4)


# draw_graph("../../paper_src/ets_update_90",
#            "ETS Mixed (90% Lookups, 10% Updates)",
#            "\"mixed=l:90.00%,=u:10.00%=4194304\"",
#            "\[", ",21,20,22,0.1,rw,1,{0,0,0}\]",
#            2**22,
#            xmax=None,
#            ymax=None,
#            legendLoc=4)


# draw_graph("../../paper_src/ets_update_99",
#            "ETS Mixed (99% Lookups, 1% Updates)",
#            "\"mixed=l:99.00%,=u:1.00%=4194304\"",
#            "\[", ",21,20,22,0.01,rw,1,{0,0,0}\]",
#            2**22,
#            xmax=None,
#            ymax=None,
#            legendLoc=2)

# #Set Size Benchmark

# for (i, legLoc) in [(3, 0),
#                     (6, 0),
#                     (9, 0),
#                     (12, 0),
#                     (15, 0),
#                     (18, 0),
#                     (21, 0),
#                     (24, 0),
#                     (27, 0)]:
#     draw_graph("../../paper_src/ets_size_" + str(i),
#                "ETS Size (Key range "+ str(int(2**i)) +", 80% Lookups, 20% Updates)",
#                "\"mixed=l:80.00%,=u:20.00%=4194304\"",
#                "\[", ","+str(i)+","+str(i-1)+",22,0.2,rw,1,{0,0,0}\]",
#                2**22,
#                xmax=None,
#                ymax=None,
#                legendLoc=0,
#                theFile="graph_src/SetSizeResults/bench_s_"+str(i)+"/ets_bench/measurements/R17.PIN_RG.sched.time")


# # Sequential performance

# import numpy as np
# N = 9
# inditer = ind = np.arange(N)  # the x locations for the groups
# width = 0.20       # the width of the bars

# plt.figure(figsize=(8,4.75))
# plt.autoscale(enable=True, tight=False)
# fig, ax = plt.subplots()

# colors = ['red','blue','yellow','green']
# patterns = ('-', '*', 'x', 'o', '+', '\\', 'O', '.')
# index=0
# rects = []
# for (theType, name, mark) in table_types_and_names:
#     times = []
#     error_mins = []
#     error_maxs = []
#     for (i, legLoc) in [(3, 0),
#                         (6, 0),
#                         (9, 0),
#                         (12, 0),
#                         (15, 0),
#                         (18, 0),
#                         (21, 0),
#                         (24, 0),
#                         (27, 0)]:
#         dataFile = "graph_src/SetSizeResults/bench_s_"+str(i)+"/ets_bench/measurements/R17.PIN_RG.sched.time"
#         with open(dataFile, 'r') as f:
#             fileContent = ("\n").join(f.readlines())
#             matchPattern = "\"mixed=l:80.00%,=u:20.00%=4194304\"\t\[" + theType + ","+str(i)+","+str(i-1)+",22,0.2,rw,1,{0,0,0}\]" + "\t1\t(.*)"
#             theMatch = re.search(matchPattern, fileContent)
#             [avg,mini,maxi] = theMatch.group(1).split()
#             avg_seconds = float(avg)/1000
#             times.append(avg_seconds)
#             min_seconds = float(mini)/1000
#             error_mins.append(avg_seconds-min_seconds)
#             max_seconds = float(maxi)/1000
#             error_maxs.append(max_seconds-avg_seconds)
#     rects.append(ax.bar(inditer, times, width,yerr=[error_mins,error_maxs], color=colors[index], edgecolor='black', hatch=patterns[index]))
#     inditer = inditer + width
#     index = index + 1

# ax.set_ylabel('Time (Seconds)')
# ax.set_xlabel('Set Size')
# ax.set_xticks(ind+width)
# ax.set_xticklabels( ('$2^{3}$', '$2^{6}$', '$2^{9}$', '$2^{12}$', '$2^{15}$', '$2^{18}$', '$2^{21}$', '$2^{24}$', '$2^{27}$') )
# ax.legend( (rects[0][0], rects[1][0], rects[2][0], rects[3][0]), ('set', 'ordset', 'AVL-CA tree', 'Treap-CA tree'),loc=0)
# plt.savefig('../../paper_src/seq_perf.pdf', bbox_inches='tight')


# # Compare to other reader writers lock

# table_types_and_names = [("\[set,21,20,22,[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?,rw", "set", 'o'),
#                          ("\[ordered_set,21,20,22,[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?,rw", "ordset", 's'),
#                          ("\[avlatree,21,20,22,[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?,rw", "AVL-CA tree", 'D'),
#                          ("\[treapatree,21,20,22,[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?,rw", "Treap-CA tree", '^'),
#                          ("plainlib.PIN_RG \[ordered_set,21,20,22,[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?,rw", "DR-MCS", '^'),
#                          ("puredel.PIN_RG \[ordered_set,21,20,22,[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?,rw", "MR-QD", '^')]


# draw_graph("../../paper_src/ets_delete_tbllock",
#            "ETS Delete",
#            "\"delete=1048576\"",
#            "", ",1,{0,0,0}\]",
#            2**20,
#            xmax=None,
#            ymax=12)

# draw_graph("../../paper_src/ets_insert_tbllock",
#            "ETS Insert",
#            "\"insert=1048576\"",
#            "", ",1,{0,0,0}\]",
#            2**20,
#            xmax=None,
#            ymax=9)

# draw_graph("../../paper_src/ets_lookup_tbllock",
#            "ETS Lookup",
#            "\"lookup=4194304\"",
#            "", ",1,{0,0,0}\]",
#            2**22,
#            xmax=None,
#            ymax=None,
#            legendLoc=2)


# draw_graph("../../paper_src/ets_update_0_tbllock",
#            "ETS Mixed (100% Updates)",
#            "\"mixed=l:0.00%,=u:100.00%=4194304\"",
#            "", ",1,{0,0,0}\]",
#            2**22,
#            xmax=None,
#            ymax=None,
#            legendLoc=1)

# draw_graph("../../paper_src/ets_update_50_tbllock",
#            "ETS Mixed (50% Lookups, 50% Updates)",
#            "\"mixed=l:50.00%,=u:50.00%=4194304\"",
#            "", ",1,{0,0,0}\]",
#            2**22,
#            xmax=None,
#            ymax=None,
#            legendLoc=4)

# draw_graph("../../paper_src/ets_update_67_tbllock",
#            "ETS Mixed (67% Lookups, 33% Updates)",
#            "\"mixed=l:67.00%,=u:33.00%=4194304\"",
#            "", ",1,{0,0,0}\]",
#            2**22,
#            xmax=None,
#            ymax=None,
#            legendLoc=4)


# draw_graph("../../paper_src/ets_update_80_tbllock",
#            "ETS Mixed (80% Lookups, 20% Updates)",
#            "\"mixed=l:80.00%,=u:20.00%=4194304\"",
#            "", ",1,{0,0,0}\]",
#            2**22,
#            xmax=None,
#            ymax=None,
#            legendLoc=4)


# draw_graph("../../paper_src/ets_update_90_tbllock",
#            "ETS Mixed (90% Lookups, 10% Updates)",
#            "\"mixed=l:90.00%,=u:10.00%=4194304\"",
#            "", ",1,{0,0,0}\]",
#            2**22,
#            xmax=None,
#            ymax=None,
#            legendLoc=4)


# draw_graph("../../paper_src/ets_update_99_tbllock",
#            "ETS Mixed (99% Lookups, 1% Updates)",
#            "\"mixed=l:99.00%,=u:1.00%=4194304\"",
#            "", ",1,{0,0,0}\]",
#            2**22,
#            xmax=None,
#            ymax=None,
#            legendLoc=2)
