import sbt._
import Keys._
import Process._
import java.io._
import Keys._
import sbt.Process._



object ScalaMapsBuild extends Build {
import sbt.Process._

 /* helpers */

  def loginfo(msg: String) = println(msg)
  
  def runsync(comm: String) {
    loginfo("running: %s".format(comm))
    comm !;
  }

  val num_of_benchmark_runs = 25 //12
  
  def benchmarkCommand(args:String) = {
    "scala -J-Xmx2048m -J-Xms2048m -J-server -cp %s:%s:%s %s".format(
      "lib/high-scale-lib.jar",
      "target/scala-2.11.1/test-classes",
      "target/scala-2.11.1/classes",
      args)
  }
  
  val plotresfile = File.createTempFile("plot", "out")
  plotresfile.deleteOnExit();
  
  val gplotflag = "--gnuplot"


  def generatePlots(cases: Seq[String], paramtups: List[(String, List[String])], input: String, output: String) = for (((pname, _), i) <- paramtups.zipWithIndex) {
    
    
    val plotfile = new java.io.File("tmp/generateplot")
    val contents = """
#!/usr/bin/gnuplot

set terminal png
set grid
set xlabel "%s"
set ylabel "time"
set out "./tmp/%s"

plot %s
""".format(
  pname.split('=')(0),
  output + i + ".png",
  (for (j <- 0 until cases.length) yield {
    """ "%s" using %d:%d title "%s" with linespoints""".format(
      input, 
      i + 1,
      paramtups.length + 1 + j,
      cases(j)
    )
  }).mkString(", ")
)
    
    val out = new java.io.FileWriter(plotfile)
    out.write(contents)
    out.close
    
    "gnuplot %s".format(plotfile) !;
  }


  /* tasks */
  
  val bench = InputKey[Unit]("bench")

  val benchTask = bench <<= inputTask { (argTask: TaskKey[Seq[String]]) =>
    (argTask, packageBin in Test) map { (args: Seq[String], pb) =>
      runsync(benchmarkCommand(args.mkString(" ")))
    }
  }

  val benchBatch = InputKey[Unit]("bench-batch")

  val benchBatchTask = benchBatch <<= inputTask { (argTask: TaskKey[Seq[String]]) =>
    (argTask, packageBin in Test) map { (args: Seq[String], pb) =>

val (params, nonparams) = args.partition(_.startsWith("-D"))
      val (other, benches) = nonparams.partition(_.startsWith("--"))
      val othermap = Map(other.map(_.split('=')).map(x => (x(0), x(1))): _*)
      
      val paramtups = for {
        p <- params.toList
        val pts = p.split('=')
        val (name, vals) = (pts(0), pts(1).split(','))
      } yield (p, vals.toList)
      
      def forparams(paramlist: List[(String, List[String])], args: List[String], accs: String)(f: (List[String], String) => Unit): Unit = paramlist match {
        case head :: tail => for (v <- head._2) forparams(tail, args ::: List(v), "%s %s=%s".format(accs, head._1.split('=')(0), v))(f)
        case Nil => f(args, accs)
      }
      
      if (othermap contains (gplotflag)) {
        val gpf = (file("tmp") / othermap(gplotflag)).asFile
        gpf.delete()
        gpf.createNewFile()
        
        forparams(paramtups, Nil, "") { (lst, s) =>
          plotresfile.delete()
          plotresfile.createNewFile()
          for (bench <- benches) {
            loginfo("running for: %s, %s".format(bench, s))
            loginfo("Command: %s %s %d 1".format(s, bench, num_of_benchmark_runs))
            val bcomm = benchmarkCommand("%s %s %d 1".format(s, bench, num_of_benchmark_runs))
            //loginfo("command: %s".format(bcomm))
            bcomm #| """sed -e s/[^0-9\t]//g """.format(bench) #>> plotresfile !;
          }
          //"cat %s".format(plotresfile.getAbsolutePath) !;
          "echo -ne %s ".format(lst.mkString(" ")) #>> gpf !;
          runsync("tools/toplot %s %s".format(plotresfile.getAbsolutePath, gpf.toString))
        }
        
        generatePlots(benches, paramtups, gpf.getAbsolutePath, othermap(gplotflag))
      } else {
        forparams(paramtups, Nil, "") { (lst, s) =>
          for (bench <- benches) {
            runsync(benchmarkCommand("%s %s %d 1".format(s, bench, num_of_benchmark_runs)))
          }
        }
      }      
    }
  } 

  
   
  //val clean = TaskKey[Unit]("clean", "Deletes files produced by the build, such as generated sources, compiled classes, and task caches.")

  override lazy val settings = super.settings ++ Seq()

  lazy val root = Project(id = "ScalaMapsBuild",
    base = file("."),
    settings = Project.defaultSettings ++ Seq(benchTask, benchBatchTask))

}
