package se.uu.collection;

import java.lang.reflect.Field;
import java.util.concurrent.atomic.AtomicLongFieldUpdater;

import sun.misc.Unsafe;

public class SeqRWLock {
	
    volatile long seqNumber = 2L;
    volatile long readIndicator = 0L;
    volatile long writeBarrier = 0L;
    
    private static final AtomicLongFieldUpdater<SeqRWLock> seqNumberUpdater =
            AtomicLongFieldUpdater.newUpdater(SeqRWLock.class, "seqNumber");
    
    private static final AtomicLongFieldUpdater<SeqRWLock> writeBarrierUpdater =
            AtomicLongFieldUpdater.newUpdater(SeqRWLock.class, "writeBarrier");
    
    private static final AtomicLongFieldUpdater<SeqRWLock> readIndicatorUpdater =
            AtomicLongFieldUpdater.newUpdater(SeqRWLock.class, "readIndicator");
    
    private static final Unsafe unsafe;

    static {
        try {
            Field theUnsafe = Unsafe.class.getDeclaredField("theUnsafe");
            theUnsafe.setAccessible(true);
            unsafe = (Unsafe) theUnsafe.get(null);
        } catch (Exception ex) { 
            throw new Error(ex);
        }
    }
    
    public void indicateReadStart(){
    	readIndicatorUpdater.getAndIncrement(this);
    }

    public void indicateReadEnd(){
    	readIndicatorUpdater.getAndDecrement(this);
    }
    
    private void waitNoOngoingRead(){
    	while(0L != readIndicator){
    		unsafe.fullFence();
    		unsafe.fullFence();
         }
    }
    
    
	public boolean tryLock() {
    	long readSeqNumber = seqNumber;
		if(writeBarrier != 0 || readIndicator != 0 || (readSeqNumber % 2) != 0){
			return false;
		}else{
			boolean success = seqNumberUpdater.compareAndSet(this, readSeqNumber, readSeqNumber + 1);
			if(success){
			    if(readIndicator != 0){
				seqNumber = readSeqNumber + 2;
				return false;
			    }else{
				return true;
			    }
			}else{
				return false;
			}
		}
	}
    
    public void lock(){
    	while(writeBarrier != 0L){
    		Thread.yield();
    	}
    	while(true){
        	long readSeqNumber = seqNumber;	
        	while((readSeqNumber % 2) != 0){
        		unsafe.fullFence();
        		unsafe.fullFence();
        		readSeqNumber = seqNumber;
        	}
        	if(seqNumberUpdater.compareAndSet(this, readSeqNumber, readSeqNumber + 1)){
        		break;
        	}
    	}
    	waitNoOngoingRead();
    }

    public void unlock(){
    	seqNumber = seqNumber + 1;
    }
    
    public boolean isWriteLocked(){
    	return (seqNumber % 2) != 0;
    }
    
    public void readLock(){
    	boolean barrierRaised = false;
    	int patience = 10000;
    	while(true){
    		indicateReadStart();
    		if(isWriteLocked()){
    			indicateReadEnd();
    			while(isWriteLocked()){
    				//Thread.yield();
            		unsafe.fullFence();
            		unsafe.fullFence();
    				if(patience == 0 && !barrierRaised){
    					writeBarrierUpdater.incrementAndGet(this);
    					barrierRaised = true;
    				}
    				patience--;
    			}
    		}else{
    			break;
    		}
    	}
		if(barrierRaised){
			writeBarrierUpdater.decrementAndGet(this);
		}
    }

    public void readUnlock(){
    	indicateReadEnd();    	
    }

	public long tryOptimisticRead() {
		long readSeqNumber = seqNumber;
		if((readSeqNumber % 2) != 0){
			return 0;
		}else{
			return readSeqNumber;
		}
	}

	public boolean validate(long optimisticReadToken) {
		unsafe.loadFence();
		long readSeqNumber = seqNumber;
		return readSeqNumber == optimisticReadToken;
	}


    
}
